unit Demo7Settings;

interface

uses
  PtSettings, Classes;

type
  //  Settings group with an integer based array property (actually property Elems)

  TIntIndexSettings = class(TSettingsGroup)
  private
    fElems: TStringList;
    function GetCount: integer;
    procedure SetCount(const Value: integer);
  protected
    //  Generate settings names
    function  ElemName(const i: integer): string;
    //  These methods do the actual work
    procedure Loaded(Storage: ISettingsStorage); override;
    procedure Saving(Storage: ISettingsStorage); override;
  public
    constructor Create(const AName: string; const AParent: TSettingsGroup);
    destructor  Destroy; override;
    //  This could just as well have been an array property
    property  Elems : TStringList read fElems;
  published
    property  Count : integer read  GetCount write SetCount;
  end;

  //  Settings group with an enumerated array property

  TEnum = (eFirst, eSecond, eThird, eFourth);

  TEnumIndexSettings = class(TSettingsGroup)
  private
    fElems: array [TEnum] of string;
    function GetElems(en: TEnum): string;
    procedure SetElems(en: TEnum; const Value: string);
  protected
    //  Generate settings names
    function  ElemName(const en: TEnum): string;
    //  These methods do the actual work
    procedure Loaded(Storage: ISettingsStorage); override;
    procedure Saving(Storage: ISettingsStorage); override;
  public
    //  Array properties cannot be published
    property  Elems[en:TEnum] : string read GetElems write SetElems;
  end;

  //  Settings object

  TSettings = class(TSettingsGroup)
  private
    fIntIndex: TIntIndexSettings;
    fEnumIndex: TEnumIndexSettings;
  public
    constructor Create(const AName: string; const AParent: TSettingsGroup = nil);
    property  IntIndex  : TIntIndexSettings   read fIntIndex;
    property  EnumIndex : TEnumIndexSettings  read fEnumIndex;
  end;

var
  Settings: TSettings;
  Storage: ISettingsStorage;

implementation

uses
  PtSettingsIni, Demos, SysUtils, Windows;

procedure Initialize;
begin
  Settings := TSettings.Create( 'Demo7', nil);
  Storage := TIniSettingsStorage.Create(FileName('.ini'));
//  Storage := TRegSettingsStorage.Create( HKEY_LOCAL_MACHINE, '\SOFTWARE\PtSettings' );
  Storage.LoadAll( Settings );
end;

procedure Finalize;
begin
  Storage.SaveAll( Settings );
  Settings.Free;
  //Storage.Free;   The Storage variable is an interface
end;

{ TIntIndexSettings }

constructor TIntIndexSettings.Create(const AName: string; const AParent: TSettingsGroup);
begin
  inherited;
  fElems  := TStringList.Create;
end;

destructor TIntIndexSettings.Destroy;
begin
  fElems.Free;
  inherited;
end;

function TIntIndexSettings.ElemName(const i: integer): string;
begin
  Result  := 'Elem_'+IntToStr(i)
end;

function TIntIndexSettings.GetCount: integer;
begin
  Result  := fElems.Count;
end;

procedure TIntIndexSettings.Loaded(Storage: ISettingsStorage);
var
  i : integer;
  Store : ISettingsStorageRead;
begin
  inherited;
  //  Get interface for explicitly reading storage values
{$IFDEF VER120} //  If Delphi 4 then use...
  Store := Storage as ISettingsStorageRead;
{$ELSE}         //  otherwise use...
  if  Supports( Storage, ISettingsStorageRead, Store ) then
{$ENDIF}
    //  Read each setting into the array
    for i := 0 to Count-1 do
      Elems[i]  := Store.ReadString( ElemName(i) )
end;

procedure TIntIndexSettings.Saving(Storage: ISettingsStorage);
var
  i : integer;
  Store : ISettingsStorageWrite;
begin
  inherited;
  //  Get interface for explicitly writing storage values
{$IFDEF VER120} //  If Delphi 4 then use...
  Store := Storage as ISettingsStorageWrite;
{$ELSE}         //  otherwise use...
  if  Supports( Storage, ISettingsStorageWrite, Store ) then
{$ENDIF}
    //  Write each setting in the array
    for i := 0 to Count-1 do
      Store.WriteString( ElemName(i), Elems[i] );
end;

procedure TIntIndexSettings.SetCount(const Value: integer);
var
  i : integer;
begin
  //  Ideally this would be :
  //    fElems.Count := Value;
  for i := fElems.Count to Value-1  do
    fElems.Add('')
end;

{ TEnumIndexSettings }

function TEnumIndexSettings.ElemName(const en: TEnum): string;
begin
  Result  := 'Enum_'+IntToStr( integer(en) )
end;

function TEnumIndexSettings.GetElems(en: TEnum): string;
begin
  Result  := fElems[en]
end;

procedure TEnumIndexSettings.Loaded(Storage: ISettingsStorage);
var
  en : TEnum;
  Store : ISettingsStorageRead;
begin
  inherited;
  //  Get interface used for explicitly reading storage values
{$IFDEF VER120} //  If Delphi 4 then use...
  Store := Storage as ISettingsStorageRead;
{$ELSE}         //  otherwise use...
  if  Supports( Storage, ISettingsStorageRead, Store ) then
{$ENDIF}
    //  Read each setting into the array
    for en := Low(TEnum) to High(TEnum) do
      Elems[en] := Store.ReadString( ElemName(en) )
end;

procedure TEnumIndexSettings.Saving(Storage: ISettingsStorage);
var
  en : TEnum;
  Store : ISettingsStorageWrite;
begin
  inherited;
  //  Get interface used for explicitly writing storage values
{$IFDEF VER120} //  If Delphi 4 then use...
  Store := Storage as ISettingsStorageWrite;
{$ELSE}         //  otherwise use...
  if  Supports( Storage, ISettingsStorageWrite, Store ) then
{$ENDIF}
    //  Write each setting in the array
    for en := Low(TEnum) to High(TEnum) do
      Store.WriteString( ElemName(en), Elems[en] );
end;

procedure TEnumIndexSettings.SetElems(en: TEnum; const Value: string);
begin
  fElems[en]  := Value;
end;

{ TSettings }

constructor TSettings.Create(const AName: string; const AParent: TSettingsGroup);
begin
  inherited;
  fIntIndex := TIntIndexSettings.Create('IntIndex', Self);
  fEnumIndex := TEnumIndexSettings.Create('EnumIndex', Self);
end;

initialization
  Initialize;

finalization
  Finalize;

end.

