unit MyXML;
//  Versions:     Delphi 6
//  Created by:   Peter Thnell
//  Contact:      thonellp@hotmail.com
//  When:         2002-08

//  I made the unit MyXML in less than a day only so that XML facilities would
//  be available for PtSettings in Delphi Personal and Delphi Professional.
//  MyXML is currently NOT supported.  So I apologise for any bugs or lack of
//  functionality you might find in it.  It seems to do the job alright though.

interface

uses
  Classes, SysUtils;

type
  //  Interfaces

  IMyXMLNodes = interface;
  IMyXMLAttributes = interface;

  IMyXMLBase = interface
  ['{2B79D884-E9D8-4829-BC36-BFDECF1D3828}']
    function GetAsString: string;
    procedure SetAsString(const Value: string);
  //  public
    property  AsString  : string  read GetAsString  write SetAsString;
  end;

  IMyXMLNode = interface(IMyXMLBase)
  ['{9590D656-7028-4343-A5F1-D01ED8A1C206}']
    function GetChildren: IMyXMLNodes;
    function GetName: string;
    function GetParent: IMyXMLNode;
    function GetAttributes: IMyXMLAttributes;
    procedure SetName(const Value: string);
  //  public
    property  Parent      : IMyXMLNode        read GetParent;
    property  Name        : string            read GetName        write SetName;
    property  Attributes  : IMyXMLAttributes  read GetAttributes;
    property  Children    : IMyXMLNodes       read GetChildren;
  end;

  IMyXMLNodes = interface(IMyXMLBase)
  ['{72B966A7-0FEB-4EFB-B4EE-7DE1BE22771D}']
    function GetOwner: IMyXMLNode;
    function GetCount: integer;
    function GetXMLNode(Index: Variant): IMyXMLNode;
  //  public
    property  Owner : IMyXMLNode  read GetOwner;
    property  Count : integer read GetCount;
    property  Nodes[Index:Variant]  : IMyXMLNode  read GetXMLNode; default;
    function  NewNode(const Name: string): IMyXMLNode;
    procedure Remove(const NodeName: string);
    procedure Clear;
  end;

  IMyXMLAttributes = interface(IMyXMLBase)
  ['{33DB2F0F-3A22-4586-AA1A-EE013FC02A21}']
    function GetOwner: IMyXMLNode;
    function GetCount: integer;
    function GetAsString: string;
    procedure SetAsString(const Value: string);
    function GetAttributes(Index: Variant): Variant;
    procedure SetAttributes(Index: Variant; const Value: Variant);
    function GetNames(Index: integer): string;
  //  public
    property  Owner : IMyXMLNode  read GetOwner;
    property  Count : integer read GetCount;
    property  Attributes[Index:Variant] : Variant read GetAttributes  write SetAttributes; default;
    property  Names[Index:integer]      : string  read GetNames;
    procedure Remove(const Index:Variant);
    procedure Clear;
  end;

  IMyXMLDeclaration = interface(IMyXMLBase)
  ['{34ED4797-32BC-46A8-A170-2FEB065A4FAC}']
    function GetVersion: string;
    procedure SetVersion(const Value: string);
    function GetStandAlone: Boolean;
    procedure SetStandAlone(const Value: Boolean);
  //  public
    property  Version     : string  read GetVersion     write SetVersion;
    property  StandAlone  : Boolean read GetStandAlone  write SetStandAlone;
  end;

  IMyXMLDocument = interface(IMyXMLBase)
  ['{97181841-2A16-401C-9F66-BA648653A761}']
    function GetNode: IMyXMLNode;
    function GetDeclaration: IMyXMLDeclaration;
    procedure SetNode(const Value: IMyXMLNode);
  //  public
    property  Declaration : IMyXMLDeclaration read GetDeclaration;
    property  Node        : IMyXMLNode        read GetNode    write SetNode;
    procedure LoadFromFile(const FileName: string);
    procedure SaveToFile(const FileName: string);
  end;

  //  Implementation

  TMyXMLNodes = class;

  TMyXMLNode = class(TInterfacedObject, IMyXMLNode)
  private
    fName : string;
    fParent : IMyXMLNode;
    fAttributes : IMyXMLAttributes;
    fChildren : IMyXMLNodes;
    function GetChildren: IMyXMLNodes;
    function GetName: string;
    function GetParent: IMyXMLNode;
    function GetAttributes: IMyXMLAttributes;
    procedure SetName(const Value: string);
  protected
    function GetAsString: string; virtual;
    procedure SetAsString(const Value: string); virtual;
  public
    constructor Create(const AName: string; const AParent: IMyXMLNode = nil);
    property  Parent      : IMyXMLNode        read GetParent;
    property  Name        : string            read GetName        write SetName;
    property  Attributes  : IMyXMLAttributes  read GetAttributes;
    property  Children    : IMyXMLNodes       read GetChildren;
  end;

  TMyXMLNodes = class(TInterfacedObject, IMyXMLNodes)
  private
    fOwner  : IMyXMLNode;
    fNodes  : array of IMyXMLNode;
    function GetCount: integer;
    function GetXMLNode(Index: Variant): IMyXMLNode;
    function GetOwner: IMyXMLNode;
    function GetAsString: string;
    procedure SetAsString(const Value: string);
  protected
    function  IndexOfName(const NodeName: string): integer;
    function  IndexOfVariant(const Index: Variant): integer;
  public
    constructor Create(const AOwner: IMyXMLNode);
    destructor  Destroy; override;
    property  Owner : IMyXMLNode  read GetOwner;
    property  Count : integer read GetCount;
    property  Nodes[Index:Variant]  : IMyXMLNode  read GetXMLNode; default;
    function  NewNode(const Name: string): IMyXMLNode;
    procedure Remove(const NodeName: string);
    procedure Clear;
  end;

  TAttrRec = record
    Name  : string;
    Value : Variant;
  end;

  TMyXMLAttributes = class(TInterfacedObject, IMyXMLAttributes)
  private
    fOwner: IMyXMLNode;
    fAttrs: array of TAttrRec;
    function GetAsString: string;
    procedure SetAsString(const Value: string);
    function GetAttributes(Index: Variant): Variant;
    function GetCount: integer;
    function GetOwner: IMyXMLNode;
    procedure SetAttributes(Index: Variant; const Value: Variant);
    function GetNames(Index: integer): string;
  protected
    function  IndexOfName(const NodeName: string): integer;
    function  IndexOfVariant(const Index: Variant): integer;
  public
    constructor Create(const AOwner: IMyXMLNode);
    property  Owner : IMyXMLNode  read GetOwner;
    property  Count : integer read GetCount;
    property  Attributes[Index:Variant] : Variant read GetAttributes write SetAttributes; default;
    property  Names[Index:integer]      : string  read GetNames;
    procedure Remove(const Index:Variant);
    procedure Clear;
  end;

  TMyXMLDeclaration = class(TInterfacedObject, IMyXMLDeclaration)
  private
    fVersion: string;
    fStandAlone: Boolean;
    function GetVersion: string;
    procedure SetVersion(const Value: string);
    function GetStandAlone: Boolean;
    procedure SetStandAlone(const Value: Boolean);
  protected
    function GetAsString: string;
    procedure SetAsString(const Value: string);
  public
    constructor Create;
    property  Version     : string  read GetVersion     write SetVersion;
    property  StandAlone  : Boolean read GetStandAlone  write SetStandAlone;
  end;

  TMyXMLDocument = class(TInterfacedObject, IMyXMLDocument)
  private
    fNode: IMyXMLNode;
    fDeclaration: IMyXMLDeclaration;
    function GetNode: IMyXMLNode;
    function GetDeclaration: IMyXMLDeclaration;
    procedure SetNode(const Value: IMyXMLNode);
  protected
    function GetAsString: string; virtual;
    procedure SetAsString(const Value: string); virtual;
  public
    constructor Create;
    property  Declaration : IMyXMLDeclaration read GetDeclaration;
    property  Node        : IMyXMLNode        read GetNode      write SetNode;
    property  AsString    : string            read GetAsString  write SetAsString;
    procedure LoadFromFile(const FileName: string);
    procedure SaveToFile(const FileName: string);
  end;

  EMyXML = class(Exception)
  end;

function  NewXMLDocument: IMyXMLDocument;

implementation

uses
  Variants;

function  NewXMLDocument: IMyXMLDocument;
begin
  Result  := TMyXMLDocument.Create;
end;

//  Conversion of values between raw strings and XML

type
  TXMLAlias = record
    XMLStr, StdStr: string;
  end;

const
  XMLAliasCount = 5;
  XMLAlias : array [1..XMLAliasCount] of TXMLAlias =
    (
     (XMLStr : '&amp;';  StdStr : '&'),
     (XMLStr : '&lt;';   StdStr : '<'),
     (XMLStr : '&gt;';   StdStr : '>'),
     (XMLStr : '&quot;'; StdStr : '"'),
     (XMLStr : '&apos;'; StdStr : '''')
    );

function  ReplaceStrs(Str: string; const FromStr, ToStr: string): string;
var
  p, len  : integer;
begin
  Result  := '';
  len     := Length(FromStr);
  repeat
    p := Pos( FromStr, Str );
    if  p = 0 then
      break;
    Result  := Result + Copy( Str, 1, p-1 ) + ToStr;
    Str     := Copy( Str, p+len, MaxInt );
  until false;
  Result  := Result+Str;
end;

function  StrToXML(Str: string): string;
var
  i : integer;
begin
  for i := 1 to XMLAliasCount do
    Str := ReplaceStrs( Str, XMLAlias[i].StdStr, XMLAlias[i].XMLStr );
  Result  := Str;
end;

function  XMLToStr(Str: string): string;
var
  i : integer;
begin
  for i := 1 to XMLAliasCount do
    Str := ReplaceStrs( Str, XMLAlias[i].XMLStr, XMLAlias[i].StdStr );
  Result  := Str;
end;

//  Parser

type
  TMyXMLParser = class
  protected
    i     : integer;
    Text  : string;
    procedure SetText(const AText: string);
    procedure Check(const Result: Boolean);
    //  Get tokens etc.
    procedure AbsorbWhiteSpace;
    function  Absorb(const Expect: string): Boolean;
    function  IsAt(const Expect: string): Boolean;
    function  ReadName: string;
    function  ReadValue: string;
  public
    procedure ParseDoc(const Doc: IMyXMLDocument; const AText: string = '');
    procedure ParseDecl(const Ver: IMyXMLDeclaration; const AText: string = '');
    procedure ParseNode(const Node: IMyXMLNode; const AText: string = '');
    procedure ParseAttrs(const Attrs: IMyXMLAttributes; const AText: string = '');
    procedure ParseNodes(const Nodes: IMyXMLNodes; const AText: string = '');
//    procedure Parse(const : IMyXML; const AText: string = '');
  end;

{ TMyXMLParser }

function TMyXMLParser.Absorb(const Expect: string): Boolean;
begin
  Result  := IsAt( Expect );
  if  Result  then
    inc( i, Length( Expect ) )
end;

procedure TMyXMLParser.AbsorbWhiteSpace;
begin
  while Text[i] in [' ', #9]  do
    inc(i)
end;

procedure TMyXMLParser.Check(const Result: Boolean);
begin
  if  not Result  then
    raise EMyXML.Create('Error parsing XML');
end;

function TMyXMLParser.IsAt(const Expect: string): Boolean;
begin
  AbsorbWhiteSpace;
  Result  := SameText( Expect, Copy(Text, i, Length(Expect)) );
end;

procedure TMyXMLParser.ParseAttrs(const Attrs: IMyXMLAttributes; const AText: string);
var
  AttName, AttValue : string;
begin
  repeat
    AttName   := ReadName;
    if  AttName = ''  then
      break;
    AttValue  := ReadValue;
    Attrs[ AttName ]  := AttValue;
  until false;
end;

procedure TMyXMLParser.ParseDecl(const Ver: IMyXMLDeclaration; const AText: string);

  function  StrToBool(const Str: string): Boolean;
  begin
    Result  := SameText(Str, 'yes')
  end;

begin
  SetText( AText );
  Check(Absorb('<?xml'));
  if  Absorb('version') then
    Ver.Version     := ReadValue;
  if  Absorb('standalone') then
    Ver.StandAlone  := StrToBool( ReadValue );
  Check(Absorb('?>'));
end;

procedure TMyXMLParser.ParseDoc(const Doc: IMyXMLDocument; const AText: string);
begin
  SetText( AText );
  if  IsAt('<?')  then
    ParseDecl( Doc.Declaration );
  Doc.Node  := TMyXMLNode.Create('');
  ParseNode( Doc.Node );
end;

procedure TMyXMLParser.ParseNode(const Node: IMyXMLNode; const AText: string);
begin
  SetText( AText );
  Check(Absorb('<'));
  Node.Name := ReadName;
  ParseAttrs( Node.Attributes );
  if  Absorb('>') then
    begin
      ParseNodes( Node.Children );
      Check(Absorb('</'+Node.Name+'>'));
    end
  else
    Check(Absorb('/>'));
end;

procedure TMyXMLParser.ParseNodes(const Nodes: IMyXMLNodes; const AText: string);
begin
  SetText( AText );
  while not IsAt('</')  do
    ParseNode( Nodes.NewNode('') );
end;

function TMyXMLParser.ReadName: string;
var
  j : integer;
begin
  AbsorbWhiteSpace;
  j := i;
  while not (Text[j] in [' ','/','<','>','='])  do
    inc(j);
  Result  := Copy( Text, i, j-i );
  i := j;
end;

function TMyXMLParser.ReadValue: string;
var
  j   : integer;
  Qt  : char;
begin
  Qt  := '''';
  if  Absorb('="')  then
    Qt  := '"'
  else
    Check(Absorb('='''));
  j := i;
  while not (Text[j] in [Qt]) do
    begin
      inc( j );
    end;
  Result  := Copy( Text, i, j-i );
  i := j;
  Check(Absorb(Qt));
  Result  := XMLToStr( Result )
end;

procedure TMyXMLParser.SetText(const AText: string);
begin
  if  AText <> '' then
    begin
      Text  := AText;
      i     := 1;
    end;
end;

{ TMyXMLNode }

constructor TMyXMLNode.Create(const AName: string; const AParent: IMyXMLNode);
begin
  inherited Create;
  fName       := AName;
  fAttributes := TMyXMLAttributes.Create(Self);
end;

function TMyXMLNode.GetAsString: string;
begin
  if  Children.Count = 0  then
    Result  := '<'+Name+' '+Attributes.AsString+'/>'
  else
    Result  := '<'+Name+' '+Attributes.AsString+'>'+Children.AsString+'</'+Name+'>';
end;

function TMyXMLNode.GetAttributes: IMyXMLAttributes;
begin
  Result  := fAttributes;
end;

function TMyXMLNode.GetChildren: IMyXMLNodes;
begin
  if  fChildren = nil then
    fChildren   := TMyXMLNodes.Create(Self);
  Result  := fChildren;
end;

function TMyXMLNode.GetName: string;
begin
  Result  := fName;
end;

function TMyXMLNode.GetParent: IMyXMLNode;
begin
  Result  := fParent
end;

procedure TMyXMLNode.SetAsString(const Value: string);
var
  Parser  : TMyXMLParser;
begin
  Parser  := TMyXMLParser.Create;
  try
    Parser.ParseNode( Self, Value );
  finally
    Parser.Free;
  end;
end;

procedure TMyXMLNode.SetName(const Value: string);
begin
  fName := Value
end;

{ TMyXMLNodes }

procedure TMyXMLNodes.Clear;
begin
  fNodes  := nil;
end;

constructor TMyXMLNodes.Create(const AOwner: IMyXMLNode);
begin
  inherited Create;
  fOwner  := AOwner;
end;

destructor TMyXMLNodes.Destroy;
begin
  inherited;
end;

function TMyXMLNodes.GetAsString: string;
var
  i : integer;
begin
  Result  := '';
  for i := 0 to Count-1 do
    Result  := Result+Nodes[i].AsString;
end;

function TMyXMLNodes.GetCount: integer;
begin
  Result  := Length(fNodes)
end;

function TMyXMLNodes.GetOwner: IMyXMLNode;
begin
  Result  := fOwner
end;

function TMyXMLNodes.GetXMLNode(Index: Variant): IMyXMLNode;
var
  i : integer;
begin
  i := IndexOfVariant( Index );
  if  i < 0 then
    Result  := nil
  else
    Result  := fNodes[ i ];
end;

function TMyXMLNodes.IndexOfName(const NodeName: string): integer;
begin
  Result  := Count-1;
  while Result >= 0 do
    if  SameText( fNodes[Result].Name, NodeName ) then
      break
    else
      dec( Result );
end;

function TMyXMLNodes.IndexOfVariant(const Index: Variant): integer;
begin
  case  VarType(Index)  of
    varSmallint, varInteger, varByte, varWord, varLongWord, varInt64 :
      Result  := Index;
    varOleStr, varStrArg, varString :
      Result  := IndexOfName( Index )
  else
    raise EMyXML.Create('Do not understand index of child node');
  end;
end;

function TMyXMLNodes.NewNode(const Name: string): IMyXMLNode;
begin
  Result  := TMyXMLNode.Create(Name, Owner);
  SetLength( fNodes, Count+1 );
  fNodes[Count-1] := Result;
end;

procedure TMyXMLNodes.Remove(const NodeName: string);
var
  i : integer;
begin
  i := IndexOfName( NodeName );
  if  i < 0 then
    raise EMyXML.Create('Trying to remove node '''+NodeName+''' which is not a child of '+Owner.Name);
  fNodes[i] := fNodes[Count-1];
  SetLength( fNodes, Count-1 );
end;

procedure TMyXMLNodes.SetAsString(const Value: string);
var
  Parser  : TMyXMLParser;
begin
  Parser  := TMyXMLParser.Create;
  try
    Parser.ParseNodes( Self, Value );
  finally
    Parser.Free;
  end;
end;

{ TMyXMLAttributes }

procedure TMyXMLAttributes.Clear;
begin
  fAttrs  := nil;
end;

constructor TMyXMLAttributes.Create(const AOwner: IMyXMLNode);
begin
  inherited Create;
  fOwner  := AOwner;
end;

function TMyXMLAttributes.GetAsString: string;
var
  i : integer;
begin
  Result  := '';
  for i := 0 to Count-1 do
    Result  := Result+fAttrs[i].Name+'="'+StrToXML(VarToStr(fAttrs[i].Value))+'" '
end;

function TMyXMLAttributes.GetAttributes(Index: Variant): Variant;
var
  i : integer;
begin
  i := IndexOfVariant( Index );
  if  i < 0 then
    raise EMyXML.Create('Attribute'''+Index+''' not found in XML node '+Owner.Name);
  Result  := fAttrs[i].Value
end;

function TMyXMLAttributes.GetCount: integer;
begin
  Result  := Length( fAttrs );
end;

function TMyXMLAttributes.GetNames(Index: integer): string;
begin
  Result  := fAttrs[Index].Name
end;

function TMyXMLAttributes.GetOwner: IMyXMLNode;
begin
  Result  := fOwner
end;

function TMyXMLAttributes.IndexOfName(const NodeName: string): integer;
begin
  Result  := Count-1;
  while Result >= 0 do
    if  SameText( fAttrs[Result].Name, NodeName ) then
      break
    else
      dec( Result );
end;

function TMyXMLAttributes.IndexOfVariant(const Index: Variant): integer;
begin
  case  VarType(Index)  of
    varSmallint, varInteger, varByte, varWord, varLongWord, varInt64 :
      Result  := Index;
    varOleStr, varStrArg, varString :
      Result  := IndexOfName( Index )
  else
    raise EMyXML.Create('Do not understand index of attribute');
  end;
end;

procedure TMyXMLAttributes.Remove(const Index: Variant);
var
  i : integer;
begin
  i := IndexOfVariant( Index );
  fAttrs[i] := fAttrs[Count-1];
  SetLength( fAttrs, Count-1 );
end;

procedure TMyXMLAttributes.SetAsString(const Value: string);
var
  Parser  : TMyXMLParser;
begin
  Parser  := TMyXMLParser.Create;
  try
    Parser.ParseAttrs( Self, Value );
  finally
    Parser.Free;
  end;
end;

procedure TMyXMLAttributes.SetAttributes(Index: Variant; const Value: Variant);
var
  i : integer;
begin
  i := IndexOfVariant( Index );
  if  i < 0 then
    begin //  Add attribute
      SetLength( fAttrs, Count+1 );
      fAttrs[Count-1].Name  := Index;
      fAttrs[Count-1].Value := Value;
    end
  else
    fAttrs[i].Value := Value;
end;

{ TMyXMLDeclaration }

constructor TMyXMLDeclaration.Create;
begin
  inherited;
  fVersion  := '1.0';
end;

function TMyXMLDeclaration.GetAsString: string;
begin
  Result  := '<?xml version="'+Version+'" standalone="';
  if  StandAlone  then
    Result  := Result + 'yes'
  else
    Result  := Result + 'no';
  Result  := Result + '" ?>'
end;

function TMyXMLDeclaration.GetStandAlone: Boolean;
begin
  Result  := fStandAlone;
end;

function TMyXMLDeclaration.GetVersion: string;
begin
  Result  := fVersion
end;

procedure TMyXMLDeclaration.SetAsString(const Value: string);
var
  Parser  : TMyXMLParser;
begin
  Parser  := TMyXMLParser.Create;
  try
    Parser.ParseDecl( Self, Value );
  finally
    Parser.Free;
  end;
end;

procedure TMyXMLDeclaration.SetStandAlone(const Value: Boolean);
begin
  fStandAlone := Value;
end;

procedure TMyXMLDeclaration.SetVersion(const Value: string);
begin
  fVersion  := Value
end;

{ TMyXMLDocument }

constructor TMyXMLDocument.Create;
begin
  fDeclaration  := TMyXMLDeclaration.Create;
end;

function TMyXMLDocument.GetAsString: string;
begin
  Result  := Declaration.AsString + Node.AsString
end;

function TMyXMLDocument.GetDeclaration: IMyXMLDeclaration;
begin
  Result  := fDeclaration
end;

function TMyXMLDocument.GetNode: IMyXMLNode;
begin
  Result  := fNode;
end;

procedure TMyXMLDocument.LoadFromFile(const FileName: string);
var
  fle : Text;
  Str, XML  : string;
begin
  AssignFile( fle, FileName );
  try
    Reset( fle );
    XML := '';
    while  not Eof( fle ) do
      begin
        ReadLn( fle, Str );
        XML := XML+' '+Str;
      end;
    AsString  := XML;
  finally
    CloseFile( fle );
  end;
end;

procedure TMyXMLDocument.SaveToFile(const FileName: string);
var
  fle : Text;
begin
  AssignFile( fle, FileName );
  try
    Rewrite( fle );
    Write( fle, AsString );
  finally
    CloseFile( fle );
  end;
end;

procedure TMyXMLDocument.SetAsString(const Value: string);
var
  Parser  : TMyXMLParser;
begin
  Parser  := TMyXMLParser.Create;
  try
    Parser.ParseDoc( Self, Value );
  finally
    Parser.Free;
  end;
end;

procedure TMyXMLDocument.SetNode(const Value: IMyXMLNode);
begin
  fNode   := Value;
end;

end.
