unit PtSettings;
//  Versions:     Delphi 6
//  Created by:   Peter Thnell
//  Contact:      thonellp@hotmail.com
//  When:         2002-08 - 2002-10
//  Copyright:    Peter Thnell
//
//  Purpose of unit:
//    The best way to get an example of how to use this unit is to use the
//    "Settings unit" wizard in PtSettingsExpert.bpl (a Delphi plugin) to create
//    a settings unit, or to look at the demo applications.
//    Once you have a settings unit, you can:
//    - access the settings very easily
//    - selectively load, save and delete settings any time in any of the
//      formats: .INI file, XML, registry or any other by implementing the
//      ISettingsStorage interface.
//  Usage:
//    - Settings properties that are to be saved in storage must be in the
//    published section of TSettingsGroup decendants. (ex. TSettingsRect)
//    - TSettingsGroup properties:
//      - GroupName:  A path name for that particular settings group.
//      - Parent:     Used for building structures of settings groups.
//    - Load, save and delete settings by calling implementations of
//    ISettingsStorage.  The methods are:
//      - Load()      Fetch settings for a selected group.
//      - Save()      Save settings for a selected group.
//      - LoadAll()   Fetch settings for a selected group and all its children.
//      - SaveAll()   Save settings for a selected group and all its children.
//      - DeleteAll() Delete settings for a group and all its children.  Setting
//                    the parameter to nil will delete the storage itself.

interface

uses
  Classes, Contnrs, SysUtils;

type
  ISettingsStorage = interface;

  //  Settings group

{$M+}
  TSettingsGroup = class
  protected
    procedure ClearSubGroups;
    procedure Loading(Storage: ISettingsStorage); virtual;
    procedure Loaded(Storage: ISettingsStorage); virtual;
    procedure Saving(Storage: ISettingsStorage); virtual;
    procedure Saved(Storage: ISettingsStorage); virtual;
  public
    constructor Create(const AName: string; const AParent: TSettingsGroup = nil);
    destructor  Destroy; override;
    property  Parent    : TSettingsGroup  read fParent    write SetParent;
    property  GroupName : string          read fGroupName write fGroupName;
    property  GroupPath : string          read GetGroupPath;
    //  Sub-node handling
    property  SubGroupCount  : integer    read GetSubGroupCount;
    property  SubGroups[i:integer] : TSettingsGroup read GetSubGroups;
    procedure AddSubGroup(const SubGroup: TSettingsGroup);
    procedure RemoveSubGroup(const SubGroup: TSettingsGroup);
    function  SubGroupByName(const SubGroupName: string): TSettingsGroup;
  end;
{$M-}

  //  Settings Storage interfaces

  ISettingsStorage = interface
  ['{63B0EFFF-B322-40B7-B0E2-8CF092E1486F}']
    procedure Load(const Group: TSettingsGroup);
    procedure Save(const Group: TSettingsGroup);
    procedure LoadAll(const Group: TSettingsGroup);
    procedure SaveAll(const Group: TSettingsGroup);
    procedure DeleteAll(const Group: TSettingsGroup);
  end;

  ISettingsStorageRead = interface
  ['{68E29FC6-4DEB-4A17-9D63-4039F389B6D6}']
    function  ReadInt(const Name: string): integer;
    function  ReadString(const Name: string): string;
    function  ReadWideString(const Name: string): WideString;
    function  ReadFloat(const Name: string): double;
    function  ReadInt64(const Name: string): Int64;
  end;

  ISettingsStorageWrite = interface
  ['{00501459-987E-4696-AB97-A5A32DF2AF98}']
    procedure WriteInt(const Name: string; const Value: integer);
    procedure WriteString(const Name: string; const Value: string);
    procedure WriteWideString(const Name: string; const Value: WideString);
    procedure WriteFloat(const Name: string; const Value: double);
    procedure WriteInt64(const Name: string; const Value: Int64);
  end;

  ISettingsStorageDelete = interface
  ['{6E4023B8-5415-438F-9CA4-08E32A33029D}']
    procedure DeleteSetting(const Name: string);
  end;

  ISettingsStorageQuery = interface
  ['{F953A984-1C22-431A-A216-AF2BD3D762AA}']
    procedure GetSettingNames(const SettingNames: TStrings);
    procedure GetGroupNames(const GroupNames: TStrings);
  end;

  //  Settings Storage implementations

  TSettingsStorage = class(TInterfacedObject, ISettingsStorage)
  protected
    //  Load values
    procedure ReadValue(const Group: TSettingsGroup; const Name: string);
    //  IReadSettingsStorage
    function  ReadInt(const Name: string): integer; virtual;
    function  ReadString(const Name: string): string; virtual;
    function  ReadWideString(const Name: string): WideString; virtual;
    function  ReadFloat(const Name: string): double; virtual;
    function  ReadInt64(const Name: string): Int64; virtual;
    //  Save values
    procedure WriteValue(const Group: TSettingsGroup; const Name: string);
    //  IWriteSettingsStorage
    procedure WriteInt(const Name: string; const Value: integer); virtual;
    procedure WriteString(const Name: string; const Value: string); virtual;
    procedure WriteWideString(const Name: string; const Value: WideString); virtual;
    procedure WriteFloat(const Name: string; const Value: double); virtual;
    procedure WriteInt64(const Name: string; const Value: Int64); virtual;
    //  IDeleteSettingsStorage
    procedure DeleteSetting(const Name: string); virtual;
    //  IQuerySettingsStorage
    procedure GetSettingNames(const SettingNames: TStrings); virtual;
    procedure GetGroupNames(const GroupNames: TStrings); virtual;
  public
    property  PreferStrings : Boolean read fPreferStrings write fPreferStrings;
    //  ISettingsStorage
    procedure Load(const Group: TSettingsGroup); virtual;
    procedure Save(const Group: TSettingsGroup); virtual;
    procedure LoadAll(const Group: TSettingsGroup); virtual;
    procedure SaveAll(const Group: TSettingsGroup); virtual;
    procedure DeleteAll(const Group: TSettingsGroup); virtual;
  end;

  //  Exception type

  ESettings = class(Exception)
  end;

const
  SSettingsDivider    = '\';

implementation

end.
